/*----------------------------------------------------------------------------

   libtunepimp -- The MusicBrainz tagging library.  
                  Let a thousand taggers bloom!
   
   Copyright (C) 2005 Lukas Lalinsky
   
   This file is part of libtunepimp.

   libtunepimp is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   libtunepimp is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with libtunepimp; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   $Id: mp4.cpp 8497 2006-09-29 21:26:44Z luks $

----------------------------------------------------------------------------*/

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <mp4v2/mp4v2.h>
#include "metadata.h"
#include "plugin.h"
#ifndef WIN32
#include "../../lib/utf8/utf8util.h"
#endif

using namespace std;

#ifndef WIN32
#define initPlugin mp4InitPlugin
#endif

#define PLUGIN_VERSION "1.0.0"
#define PLUGIN_NAME    "MP4 metadata reader/writer"

static const char *formats[][2] = {
  { ".aac", "AAC/MP4" },
  { ".mp4", "MP4" },
  { ".m4a", "MP4" },
  { ".m4b", "MP4" },
  { ".m4p", "MP4" },
};

#define NUM_FORMATS 5

static const char *errorString = "";

static void
mp4Shutdown()
{
}

static const char *
mp4GetVersion()
{
  return PLUGIN_VERSION;
}

static const char *
mp4GetName()
{
  return PLUGIN_NAME;
}

static int
mp4GetNumFormats(void)
{
  return NUM_FORMATS;
}

static int
mp4GetFormat(int i, char ext[TP_EXTENSION_LEN],
		char desc[TP_PLUGIN_DESC_LEN],int *functions)
{
  if (i < 0 || i >= NUM_FORMATS)
    return 0;
  
  strcpy(ext, formats[i][0]);
  strcpy(desc, formats[i][0]);
  *functions = TP_PLUGIN_FUNCTION_METADATA;

  return 1;
}

static const char *
mp4GetError()
{
  return errorString;
}

static MP4ItmfItemList*
mp4GetItmfItemValue(MP4FileHandle mp4file, const char *tagName, const char **value)
{
  MP4ItmfItemList *items = MP4ItmfGetItemsByMeaning(mp4file, "com.apple.iTunes", tagName);
  if (items != NULL) {
    if (items->size == 0) {
      MP4ItmfItemListFree(items);
      return NULL;
    }
    if (items->elements[0].dataList.size == 0) {
      MP4ItmfItemListFree(items);
      return NULL;
    }
    *value = reinterpret_cast<const char*>(items->elements[0].dataList.elements[0].value);
  }
  return items;
}

void
mp4SetItmfItemValue(MP4FileHandle mp4file, const char *tagName, const char *value, size_t size)
{
  MP4ItmfItemList *items = MP4ItmfGetItemsByMeaning(mp4file, "com.apple.iTunes", tagName);
  if (items != NULL) {
    for (size_t idx = 0; idx < items->size; idx++)
      MP4ItmfRemoveItem(mp4file, items->elements + idx);
    MP4ItmfItemListFree(items);
  }

  MP4ItmfItem *item = MP4ItmfItemAlloc("----", 1);
  item->mean = strdup("com.apple.iTunes");
  item->name = strdup(tagName);
  item->dataList.elements[0].typeCode = MP4_ITMF_BT_UTF8;
  item->dataList.elements[0].valueSize = size;
  item->dataList.elements[0].value = (uint8_t*)malloc(size);
  memcpy(item->dataList.elements[0].value, value, size);
  MP4ItmfAddItem(mp4file, item);
}

static int
mp4ReadMetadata(metadata_t *mdata, const char *fileName, int flags, const char *encoding)
{
  MP4FileHandle mp4file;
  MP4ItmfItemList *items;
  const char *value;
  
#ifndef WIN32
  mp4file = MP4Read(utf8ToEncoding(fileName, encoding).c_str());
#else  
  mp4file = MP4Read(fileName);
#endif  

  if (mp4file == MP4_INVALID_FILE_HANDLE) 
    return 0;
 
  memset(mdata, 0, sizeof(metadata_t));

  const MP4Tags *tags = MP4TagsAlloc();
  if (MP4TagsFetch(tags, mp4file)) {
    if (tags->name)
      strcpy(mdata->track, tags->name);

    if (tags->artist)
      strcpy(mdata->artist, tags->artist);

    if (tags->releaseDate)
      mdata->releaseYear = atoi(tags->releaseDate); // ignore errors, so will stop on "-" after YYYY[-MM-DD]

    if (tags->album)
      strcpy(mdata->album, tags->album);

    if (tags->track) {
      mdata->trackNum = tags->track->index;
      mdata->totalInSet = tags->track->total;
    }

    if (tags->compilation)
      mdata->variousArtist = *(tags->compilation);

    MP4TagsFree(tags);
  }

  if ((items = mp4GetItmfItemValue(mp4file, "MusicBrainz Sortname", &value)) != NULL) {
    strcpy(mdata->sortName, value);
    MP4ItmfItemListFree(items);
  }
  
  if ((items = mp4GetItmfItemValue(mp4file, "MusicBrainz Track Id", &value)) != NULL) {
    strcpy(mdata->trackId, value);
    MP4ItmfItemListFree(items);
  }
  
  if ((items = mp4GetItmfItemValue(mp4file, "MusicBrainz Album Id", &value)) != NULL) {
    strcpy(mdata->albumId, value);
    MP4ItmfItemListFree(items);
  }
  
  if ((items = mp4GetItmfItemValue(mp4file, "MusicBrainz Artist Id", &value)) != NULL) {
    strcpy(mdata->artistId, value);
    MP4ItmfItemListFree(items);
  }
  
  if ((items = mp4GetItmfItemValue(mp4file, "MusicIP PUID", &value)) != NULL) {
    strcpy(mdata->filePUID, value);
    MP4ItmfItemListFree(items);
  }
  
  if ((items = mp4GetItmfItemValue(mp4file, "MusicBrainz Album Artist Id", &value)) != NULL) {
    strcpy(mdata->albumArtistId, value);
    MP4ItmfItemListFree(items);
  }
  
  if ((items = mp4GetItmfItemValue(mp4file, "MusicBrainz Album Artist Sortname", &value)) != NULL) {
    strcpy(mdata->albumArtistSortName, value);
    MP4ItmfItemListFree(items);
  }
  
  if ((items = mp4GetItmfItemValue(mp4file, "MusicBrainz Album Artist", &value)) != NULL) {
    strcpy(mdata->albumArtist, value);
    MP4ItmfItemListFree(items);
  }
  
  if ((items = mp4GetItmfItemValue(mp4file, "MusicBrainz Album Type", &value)) != NULL) {
    mdata->albumType = convertToAlbumType(value);
    MP4ItmfItemListFree(items);
  }
  
  if ((items = mp4GetItmfItemValue(mp4file, "MusicBrainz Album Status", &value)) != NULL) {
    mdata->albumStatus = convertToAlbumStatus(value);
    MP4ItmfItemListFree(items);
  }
  
  if ((items = mp4GetItmfItemValue(mp4file, "MusicBrainz Album Release Date", &value)) != NULL) {
    int year = 0, month = 0, day = 0;
    if (sscanf(value, "%04d-%02d-%02d", &year, &month, &day) > 0) {
      mdata->releaseYear  = year;
      mdata->releaseMonth = month;
      mdata->releaseDay   = day;
    }
    MP4ItmfItemListFree(items);
  }
  
  if ((items = mp4GetItmfItemValue(mp4file, "MusicBrainz Album Release Country", &value)) != NULL) {
    strcpy(mdata->releaseCountry, value);
    MP4ItmfItemListFree(items);
  }
  
  if ((items = mp4GetItmfItemValue(mp4file, "MusicBrainz Non-Album", &value)) != NULL) {
    mdata->nonAlbum = atoi(value);
    MP4ItmfItemListFree(items);
  }
  
  u_int32_t numTracks = MP4GetNumberOfTracks(mp4file);
  for (u_int32_t i = 0; i < numTracks; i++) {
    MP4TrackId trackId = MP4FindTrackId(mp4file, i);
    const char *trackType = MP4GetTrackType(mp4file, trackId);
    if (!strcmp(trackType, MP4_AUDIO_TRACK_TYPE)) {
      MP4Duration trackDuration = MP4GetTrackDuration(mp4file, trackId); 
      mdata->duration = (unsigned long)MP4ConvertFromTrackDuration(mp4file, trackId, trackDuration, MP4_MSECS_TIME_SCALE);    
    }
  }  

  strcpy(mdata->fileFormat, fileName + strlen(fileName) - 3); 
  
  MP4Close(mp4file);
  
  return 1;
}

static int
mp4WriteMetadata(const metadata_t *mdata, const char *fileName, int flags,
		   const char *encoding)
{
  char temp[256];
  string temp2;
  MP4FileHandle mp4file;
  
#ifndef WIN32
  mp4file = MP4Modify(utf8ToEncoding(fileName, encoding).c_str());
#else  
  mp4file = MP4Modify(fileName);
#endif  

  if (mp4file == MP4_INVALID_FILE_HANDLE)
    return 0;

  if ((flags & TP_PLUGIN_FLAGS_GENERAL_CLEAR_TAGS) != 0) {
    MP4ItmfItemList* items = MP4ItmfGetItems(mp4file);
    if (items != NULL) {
      for (size_t idx = 0; idx < items->size; idx++)
        MP4ItmfRemoveItem(mp4file, items->elements + idx);
      MP4ItmfItemListFree(items);
    }
  }

  const MP4Tags* mp4tags = MP4TagsAlloc();
  if (MP4TagsFetch(mp4tags, mp4file)) {
    MP4TagsSetName(mp4tags, mdata->track);
  
    MP4TagsSetArtist(mp4tags, mdata->artist);
  
  sprintf(temp, "%04d", mdata->releaseYear);
    MP4TagsSetReleaseDate(mp4tags, temp);
  
    MP4TagsSetAlbum(mp4tags, mdata->album);
  
    MP4TagTrack tagTrack = {mdata->trackNum, mdata->totalInSet};
    MP4TagsSetTrack(mp4tags, &tagTrack);

    uint8_t compilation = mdata->variousArtist ? 1 : 0;
    MP4TagsSetCompilation(mp4tags, &compilation);
  }
  
  mp4SetItmfItemValue(mp4file, "MusicBrainz Sortname", mdata->sortName, strlen(mdata->sortName) + 1);
  
  mp4SetItmfItemValue(mp4file, "MusicBrainz Track Id", mdata->trackId, strlen(mdata->trackId) + 1);
  
  mp4SetItmfItemValue(mp4file, "MusicBrainz Album Id", mdata->albumId, strlen(mdata->albumId) + 1);
  
  mp4SetItmfItemValue(mp4file, "MusicBrainz Artist Id", mdata->artistId, strlen(mdata->artistId) + 1);

  mp4SetItmfItemValue(mp4file, "MusicIP PUID", mdata->filePUID, strlen(mdata->filePUID) + 1);

  mp4SetItmfItemValue(mp4file, "MusicBrainz Album Artist Id", mdata->albumArtistId, strlen(mdata->albumArtistId) + 1);

  mp4SetItmfItemValue(mp4file, "MusicBrainz Album Artist Sortname", mdata->albumArtistSortName, strlen(mdata->albumArtistSortName) + 1);

  mp4SetItmfItemValue(mp4file, "MusicBrainz Album Artist", mdata->albumArtist, strlen(mdata->albumArtist) + 1);

  convertFromAlbumType(mdata->albumType, temp2);
  mp4SetItmfItemValue(mp4file, "MusicBrainz Album Type", temp2.c_str(), temp2.length() + 1);
  
  convertFromAlbumStatus(mdata->albumStatus, temp2);
  mp4SetItmfItemValue(mp4file, "MusicBrainz Album Status", temp2.c_str(), temp2.length() + 1);
  
  if (mdata->releaseYear > 0) {
    if (mdata->releaseMonth > 0) {
      if (mdata->releaseDay > 0) {
        sprintf(temp, "%04d-%02d-%02d", mdata->releaseYear, mdata->releaseMonth, mdata->releaseDay);
      }
      else {
        sprintf(temp, "%04d-%02d", mdata->releaseYear, mdata->releaseMonth);
      }
    }
    else { 
      sprintf(temp, "%04d", mdata->releaseYear);
    }
  }
  else {
    strcpy(temp, "");
  }
  mp4SetItmfItemValue(mp4file, "MusicBrainz Album Release Date", temp, strlen(temp) + 1);
  
  mp4SetItmfItemValue(mp4file, "MusicBrainz Album Release Country", mdata->releaseCountry, strlen(mdata->releaseCountry) + 1);

  sprintf(temp, "%d", mdata->nonAlbum);  
  mp4SetItmfItemValue(mp4file, "MusicBrainz Non-Album", temp, strlen(temp) + 1);
  
  MP4Close(mp4file);

#ifndef WIN32
  if (!MP4Optimize(utf8ToEncoding(fileName, encoding).c_str()))
#else  
  if (!MP4Optimize(fileName))
#endif    
    return 0;
  
  return 1;
}

static unsigned long
mp4GetDuration(const char *fileName, int flags, const char *encoding)
{
  return 0;
}

static Plugin methods = {
  mp4Shutdown,
  mp4GetVersion,
  mp4GetName,
  mp4GetNumFormats,
  mp4GetFormat,
  mp4GetError,
  mp4ReadMetadata,
  mp4WriteMetadata,
  mp4GetDuration,
  NULL,
  NULL,
  NULL,
  NULL
};

extern "C" Plugin *
initPlugin()
{
  return &methods;
}

#ifdef WIN32

#include <sys/timeb.h>

extern "C" int
gettimeofday (struct timeval *t, void *foo)
{
	struct _timeb temp;
	_ftime(&temp);
	t->tv_sec = temp.time;
	t->tv_usec = temp.millitm * 1000;
	return (0);
}

#endif

